%{
/*
 * Copyright (c) 2002 Anders Magnusson. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /* This is a modified verison for c99tree; these modifications are
    Copyright (c) 2010 Tibor Palinkas.
    Original file version marking:
    scan.l,v 1.94 2010/08/11 14:08:44 ragge Exp */
#undef YY_NO_INPUT
# define LLINEDEF  __LINE__
%}


D     [0-9]
L     [a-zA-Z_]
H     [a-fA-F0-9]
E     [Ee][+-]?{D}+
P     [Pp][+-]?{D}+
FS    (f|F|l|L)?i?
IS    (u|U|l|L)*
UL    ({L}|\\u{H}{H}{H}{H}|\\U{H}{H}{H}{H}{H}{H}{H}{H})
NL    [\\][\n]
ANY   (.|{NL})

%{
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "tree.h"
#include "scope.h"
#include "cgram.h"
#include "c99tree.h"

static int is_typename();
static void control(int);
static void pragma(void);

#define CPP_IDENT 2
#define CPP_LINE 3
#define CPP_HASH 4
#define CPP_TREE 5
#define CPP_BREAK 6

#if defined(FLEX_SCANNER) && YY_FLEX_SUBMINOR_VERSION == 31
/* Hack to avoid unnecessary warnings */
FILE *yyget_in  (void);
FILE *yyget_out  (void);
int yyget_leng  (void);
char *yyget_text  (void);
void yyset_in (FILE *);
void yyset_out (FILE *);
int yyget_debug  (void);
void yyset_debug (int);
int yylex_destroy  (void);
extern int yyget_lineno (void);
extern void yyset_lineno (int);
#endif

static void lus()
{
	loc_cpy(&(yylval.string.loc), loc_current);
}

/* some of these have to be a macro so that tt_append will use the
   right __LINE__ */

/* shorthand for loc_update_end for keeping track of location */
#define lue_typ(toktyp) \
do { \
	loc_update_end(yytext); \
	lus(); \
	tt_append(toktyp, strdup(yytext)); \
	yylval.string.loc.twin_parent = loc_current->twin_parent; \
} while(0)

#define lue() lue_typ(T_TOKEN_STR)

#define lue_chr() \
do { \
	yylval.string.str = NULL; \
	loc_update_end(yytext);\
	loc_cpy(&(yylval.string.loc), loc_current); \
	tt_append(T_TOKEN_STR, strdup(yytext)); \
	yylval.string.loc.twin_parent = loc_current->twin_parent; \
} while(0)

/* same as lue, but also strdups yytext into yylval.string and saves start of string */
#define luedup() \
do { \
	yylval.string.str = strdup(loc_update_end(yytext)); \
	lus(); \
	tt_append(T_TOKEN_STR, strdup(yytext)); \
	yylval.string.loc.twin_parent = loc_current->twin_parent; \
} while(0)

/* same as lue but don't create a node in the token tree */
static void lue_silent()
{
	loc_update_end(yytext);
	lus();
}


void cunput(char c);
static void read_comment();
%}

%%
"/*" { read_comment(); }

"//"(.|{NL})*       {
		lue_silent();
		lus();
		tt_append(T_COMMENT, strdup(yytext));
	}

"__func__"       { luedup(); return(C_STRING); }
"__attribute__"  { lue();    return(C_ATTRIBUTE); }
"asm"            { lue();    return(C_ASM); }
"auto"           { luedup(); return(C_CLASS); }
"_Bool"          { luedup(); return (C_TYPE); }
"break"          { luedup(); return(C_BREAK); }
"case"           { luedup(); return(C_CASE); }
"char"           { luedup(); return (C_TYPE); }
"_Complex"       { luedup(); return (C_TYPESPEC); }
"const"          { luedup(); return (C_QUALIFIER); }
"continue"       { luedup(); return(C_CONTINUE); }
"default"        { luedup(); return(C_DEFAULT); }
"do"             { luedup(); return(C_DO); }
"double"         { luedup(); return (C_TYPE); }
"else"           { lue();    return(C_ELSE); }
"enum"           { lue();    return(C_ENUM); }
"extern"         { luedup(); return (C_CLASS); }
"__thread"       { luedup(); return (C_CLASS); }
"float"          { luedup(); return (C_TYPE); }
"for"            { luedup(); return(C_FOR); }
"goto"           { luedup(); return(C_GOTO); }
"if"             { luedup(); return(C_IF); }
"_Imaginary"     { luedup(); return (C_TYPESPEC); }
"inline"         { luedup(); return(C_INLINE); }
"int"            { luedup(); return (C_TYPE); }
"long"           { luedup(); return (C_TYPESPEC); }
"register"       { luedup(); return (C_CLASS); }
"restrict"       { luedup(); return (C_QUALIFIER); }
"return"         { luedup(); return(C_RETURN); }
"short"          { luedup(); return (C_TYPESPEC); }
"signed"         { luedup(); return (C_TYPESPEC); }
"sizeof"         { luedup(); return(C_SIZEOF); }
"typeof"         { luedup(); return(C_TYPEOF); }
"static"         { luedup(); return (C_CLASS); }
"struct"         { luedup(); return(C_STRUCT); }
"switch"         { luedup(); return(C_SWITCH); }
"typedef"        { luedup(); return (C_TYPEDEF); }
"union"          { luedup(); return(C_STRUCT); }
"unsigned"       { luedup(); return (C_TYPESPEC); }
"void"           { luedup(); return (C_TYPE); }
"volatile"       { luedup(); return (C_QUALIFIER); }
"while"          { luedup(); return(C_WHILE); }

{UL}({UL}|{D})*  { luedup(); return is_typename();}

0[xX]{H}+{IS}?   { luedup(); return(C_ICON); }
0{D}+{IS}?       { luedup(); return(C_ICON); }
{D}+{IS}?        { luedup(); return(C_ICON); }
L'(\\.|[^\\'])+' { luedup(); return(C_ICON); }
'(\\.|[^\\'])+'  { luedup(); return(C_ICON); }

{D}+{E}{FS}?             { luedup(); return(C_FCON); }
{D}*"."{D}+({E})?{FS}?   { luedup(); return(C_FCON); }
{D}+"."{D}*({E})?{FS}?   { luedup(); return(C_FCON); }
0[xX]{H}*"."{H}+{P}{FS}? { luedup(); return(C_FCON); }
0[xX]{H}+"."{P}{FS}?     { luedup(); return(C_FCON); }
0[xX]{H}+{P}{FS}?        { luedup(); return(C_FCON); }

L?\"(\\.|[^\\"]|{NL})*\"      { /*"*/ luedup(); return C_STRING; }

"..."         { lue(); return(C_ELLIPSIS); }
">>="         { luedup(); return(C_ASOP); }
"<<="         { luedup(); return(C_ASOP); }
"+="          { luedup(); return(C_ASOP); }
"-="          { luedup(); return(C_ASOP); }
"*="          { luedup(); return(C_ASOP); }
"/="          { luedup(); return(C_ASOP); }
"%="          { luedup(); return(C_ASOP); }
"&="          { luedup(); return(C_ASOP); }
"^="          { luedup(); return(C_ASOP); }
"|="          { luedup(); return(C_ASOP); }
">>"          { luedup(); return(C_SHIFTOP); }
"<<"          { luedup(); return(C_SHIFTOP); }
"++"          { luedup(); return(C_INCOP); }
"--"          { luedup(); return(C_INCOP); }
"&&"          { luedup(); return(C_ANDAND); }
"||"          { luedup(); return(C_OROR); }
"<="          { luedup(); return(C_RELOP); }
">="          { luedup(); return(C_RELOP); }
"=="          { luedup(); return(C_EQUOP); }
"!="          { luedup(); return(C_EQUOP); }
";"           { lue_chr();    return(';'); }
("{"|"<%")    { lue_chr();    return('{'); }
("}"|"%>")    { lue_chr();    return('}'); }
","           { lue_chr();    return(','); }
":"           { lue_chr();    return(':'); }
"="           { lue_chr();    return('='); }
"("           { lue_chr();    return('('); }
")"           { lue_chr();    return(')'); }
("["|"<:")    { lue_chr();    return('['); }
("]"|":>")    { lue_chr();    return(']'); }
"&"           { lue_chr();    return('&'); }
"->"          { lue(); lus(); yylval.string.str = strdup("-> binary");  return(C_STROP); }
"."           { lue(); lus(); yylval.string.str = strdup(". binary");  return(C_STROP); }
"!"           { lue(); lus(); yylval.string.str = strdup("! unary"); return(C_UNOP); }
"~"           { lue(); lus(); yylval.string.str = strdup("~ unary"); return(C_UNOP); }
"-"           { lue_chr();    return('-'); }
"+"           { lue_chr();    return('+'); }
"*"           { lue_chr();    return('*'); }
"/"           { luedup(); return(C_DIVOP); }
"%"           { luedup(); return(C_DIVOP); }
"<"           { luedup(); return(C_RELOP); }
">"           { luedup(); return(C_RELOP); }
"^"           { lue_chr();    return('^'); }
"|"           { lue_chr();    return('|'); }
"?"           { lue_chr();    return('?'); }

^#pragma[ \t]{ANY}*        { pragma(); return C_PRAGMA; }
^#ident[ \t]{ANY}*         { control(CPP_IDENT); }
^#line[ \t]{ANY}*          { control(CPP_LINE); }
^#tree[ \t]{ANY}*[\r\n]    { control(CPP_TREE); }
^#include[ \t]{ANY}*       { lue_typ(T_TOKEN_PPDIR); }
^#break[ \t]{ANY}*         { control(CPP_BREAK); }
^#{ANY}*                   { control(CPP_HASH); }

[ \t\v\f\r]      { lue(); }
"\n"             { lue(); }
.                { lue(); fprintf(stderr, "Warning: lex ignoring character '%c'\n", *yytext); /* ignore bad characters */ }

%%

#include "tree.h"

void check_switch_file(char *fn)
{
	loc_switch_file(fn, fn);
}

int
yywrap(void)
{
	if (0) unput(0); /* quiet gcc */
	return(1);
}

static void cpp_break()
{
	/* forced breakpont */
}

#define fn_cleanup(wr, str,  res) \
	do { \
		res = NULL; \
		wr = str; \
		while (*wr && *wr != '\"') \
			wr++; \
		if (*wr == 0) \
			return; \
		if (*wr++ != '\"') \
			goto bad; \
		res = wr; \
		while (*wr && *wr != '\"') \
			wr++; \
		if (*wr != '\"') \
			goto bad; \
		*wr = 0; \
	} while(0)

void
control(int t)
{
	char *wr = yytext;
	char *eptr, *fn;
	int val;

	lue_silent();

	wr++;             /* Skip initial '#' */
	switch (t) {

	case CPP_BREAK:
		cpp_break();
		return;

	case CPP_IDENT:
		return;         /* Just skip these for now. */

	case CPP_LINE:
		wr += 4;
		/* FALLTHROUGH */
	case CPP_HASH:
		val = strtol(wr, &eptr, 10);
		if (wr == eptr) /* Illegal string */
			goto bad;
		fn_cleanup(wr, eptr, fn);
		check_switch_file(fn);
		loc_current->cpp_file = loc_cpp_file_lookup(fn);
		loc_current->cpp_line = val - 1;
		loc_current->eline = val - 1;
/*		fprintf(stderr, "#HASH: %d\n", loc_current->cpp_line);*/
		break;

	case CPP_TREE:
		wr += 5;
		while(isspace(*wr)) wr++;
		if (strncmp(wr, "macro_expand_begin", 18) == 0) {
			loc_current->macro_level++;
		}
		else if (strncmp(wr, "macro_expand_end", 16) == 0) {
			loc_current->macro_level--;
		}
		else if (strncmp(wr, "offs", 4) == 0) {
			char *end;
			int b1, b2;
			b1 = strtol(wr+5, &end, 10);
			while(isspace(*end)) end++;
			b2 = strtol(end, &end, 10);
			fn_cleanup(wr, end, fn);
			check_switch_file(fn);
			loc_current->offs = loc_current->soffs = b1;
			if (b2 >= 0)
				loc_current->eoffs = b2;
		}
		loc_current->cpp_line--;
		loc_current->sline--;
		loc_current->eline--;
		break;
	}
	return;
bad:
	fprintf(stderr, "%s: illegal control", yytext);
}

/*
 * got a full pragma line.  Split it up here.
 */
static void pragma()
{
	char *arg;

	lus();
	lue();
	arg = yytext+7;

	while((*arg == ' ') || (*arg == '\t')) arg++;

	yylval.string.str = strdup(arg);
}

void cunput(char c)
{
	unput(c);
}


extern char *c_typename;
static int is_typename()
{
	const node_t *n;
	n = scope_get_typedef(yylval.string.str);
	if (n != NULL) {
		if (c_typename != NULL)
			free(c_typename);
		c_typename = yylval.string.str;
		yylval.node = Nclone(n, INV_LOC_ALL_BUT_ID);
		loc_cpy(&yylval.node->loc, loc_current);
		return C_TYPENAME;
	}
	return C_NAME;
}

static char *comment = NULL;
static int alloced = 0, used = 0;

static void cappend(char c)
{
	if (used == alloced) {
		alloced += 256;
		comment = realloc(comment, alloced);
	}
	comment[used] = c;
	used++;
}


/* read in comment as documented in flex doc */
static void read_comment()
{
	int c;
	cappend('/');
	cappend('*');
  for (;;) {
		while(((c = input()) != '*') && (c != EOF)) cappend(c);    /* eat up text of comment */
    if (c == '*') {
			cappend(c);
			while ((c = input()) == '*') cappend(c);
	    if ( c == '/' ) {
				cappend(c);
				comment[used] = '\0';
				yylval.string.str = comment;
				loc_update_end(yylval.string.str);
				comment = 0;
				alloced = 0;
				used = 0;
				tt_append(T_COMMENT, yylval.string.str);
				return;
			}
			else
				cappend(c);
	  }
    if (c == EOF) {
			fprintf(stderr, "EOF in /* */ comment");
			break;
		}
	}

	/* we never get here */
}
